#
# Package-level exports
#

from ._vars import TemplateVariables
from ._vars.loading import load_vars_from_yaml_files
from ._vars.post_processing import post_process_vars
from ._rendering import RenderConfig, RenderDirResult, render_dir



import importlib
import importlib.util
from pathlib import Path
from types import ModuleType
from sys import modules as sys_modules, path as sys_path

from ..exceptions import InvalidVarsProcessorInterface, ModuleExecutionException
from ..timing import Stage, ProgressListener, NullProgressListener
from .._logging import print_warning

from . import TemplateVariables

def post_process_vars(
        vars:                          TemplateVariables,
        vars_processor_module_locator: str,
        vars_processor_function_name:  str,
        progress_listener:             ProgressListener = NullProgressListener()
    ):
    vars_processor_module = load_module(vars_processor_module_locator)
    assert isinstance(vars_processor_module, ModuleType)

    if not hasattr(vars_processor_module, vars_processor_function_name):
        raise InvalidVarsProcessorInterface(vars_processor_module_locator, vars_processor_function_name)
    vars_processor_function = vars_processor_module.__getattribute__(vars_processor_function_name)
    if not callable(vars_processor_function):
        raise InvalidVarsProcessorInterface(vars_processor_module_locator, vars_processor_function_name)

    print(f'Post-processing template variables dictionary using {vars_processor_module.__name__}.{vars_processor_function_name}()')
    progress_listener.on_stage_started(Stage.VARS_PROCESSING)
    try:
        vars_processor_function(vars)
    except BaseException as e:
        is_keyboard_interrupt = isinstance(e, KeyboardInterrupt)
        progress_listener.on_stage_ended(Stage.VARS_PROCESSING, 0 if is_keyboard_interrupt else 1, is_keyboard_interrupt)
        raise
    else:
        progress_listener.on_stage_ended(Stage.VARS_PROCESSING, 0, False)

def load_module(module_locator: str) -> ModuleType:
    try:
        module = importlib.import_module(module_locator)
        print(f"Found Python module '{module_locator}' in the environment")
    except ModuleNotFoundError:
        module = None
        pass

    if module is None:
        module = load_module_from_file(Path(module_locator))
        print(f"Loaded Python module '{module.__name__}' from {module_locator}")

    return module

def load_module_from_file(module_path: Path) -> ModuleType:
    module_name = module_path.stem
    if not module_path.is_file():
        raise ModuleNotFoundError(f"module {module_path} not found", path=str(module_path))
    module_spec = importlib.util.spec_from_file_location(module_name, module_path)
    if module_spec is None:
        raise ImportError(f"failed to import Python module from file {module_path}", path=str(module_path))

    module = importlib.util.module_from_spec(module_spec)
    if (module_spec.name in sys_modules):
        print_warning(f"warning: Replacing existing module '{module_spec.name}' with the one loaded from {module_path}")
    sys_modules[module_spec.name] = module
    sys_path.insert(0, str(module_path.parent))

    assert module_spec.loader is not None
    try:
        module_spec.loader.exec_module(module)
    except BaseException as e:
        raise ModuleExecutionException(e)

    return module


import signal
from pathlib import Path
from typing import Any
from multiprocessing import Pool
from itertools import repeat

import glob
import yaml

from ..timing import Stage, ProgressListener, NullProgressListener
from .._logging import print_warning

from . import TemplateVariables

#
# Constants
#

_VARS_ENCODING = 'utf-8' # This correctly handles UTF-8 with or without BOM for YAML files with variables

#
# Interface
#

def load_vars_from_yaml_files(vars_files_glob: str, jobs_count: int, progress_listener: ProgressListener = NullProgressListener()) -> TemplateVariables:
    var_files_paths = glob.glob(vars_files_glob, recursive=True)
    var_files_count = len(var_files_paths)

    if var_files_count <= 0:
        print_warning(f"warning: {vars_files_glob} didn't match any files; skipping vars loading")
        return TemplateVariables()

    adjusted_jobs_count = min(var_files_count, jobs_count)
    print(f"Loading template variables from {len(var_files_paths)} files matching {vars_files_glob} in {adjusted_jobs_count} parallel jobs...")

    yaml_loader_class  = _select_yaml_loader_class()
    progress_listener.on_stage_started(Stage.VARS_LOADING)
    try:
        with Pool(adjusted_jobs_count, signal.signal, (signal.SIGINT, signal.SIG_IGN)) as process_pool:
            vars_parts = process_pool.starmap(_load_dict_from_yaml_file, zip(var_files_paths, repeat(yaml_loader_class)))

        vars = TemplateVariables()
        for vars_part in vars_parts:
            _merge_in_vars(vars, vars_part)

        progress_listener.on_stage_ended(Stage.VARS_LOADING, 0, False)

        return vars
    except BaseException as e:
        is_keyboard_interrupt = isinstance(e, KeyboardInterrupt)
        progress_listener.on_stage_ended(Stage.VARS_LOADING, 0 if is_keyboard_interrupt else 1, is_keyboard_interrupt)
        raise

def inject_service_var(vars: TemplateVariables, name: str, value: Any):
    if name in vars:
        raise ValueError(f'service variable name {name} is already used')
    vars[name] = value

#
# Service
#

def _select_yaml_loader_class() -> type:
    try:
        # Use the faster loader from LibYAML bindings
        return yaml.CLoader
    except AttributeError:
        print_warning('warning: Using the slower Python-based YAML loader')
        print('hint: install LibYAML bindings to switch to the faster C-based loader')
        return yaml.Loader

def _load_dict_from_yaml_file(yaml_file_path: Path, yaml_loader_class: type) -> dict:
    with open(yaml_file_path, 'r', encoding=_VARS_ENCODING) as yaml_file:
        return yaml.load(yaml_file, Loader=yaml_loader_class)

def _merge_in_vars(vars: TemplateVariables, new_vars: dict):
    duplicate_keys = vars.keys() & new_vars
    if len(duplicate_keys) > 0:
        first_duplicate_key = next(iter(duplicate_keys))
        raise ValueError(f'encountered duplicate variable {first_duplicate_key}')

    vars.update(new_vars)



from enum import Enum
from typing import Protocol, Callable
from timeit import default_timer

#
# Enums
#

class Stage(str, Enum):
    VARS_LOADING     = "YAML variables loading"
    VARS_PROCESSING  = "Variables post-processing"
    RENDER_POOL_INIT = "Render pool initialization"
    JINJA_RENDER     = "Jinja templates rendering"

#
# Protocols
#

class ProgressListener(Protocol):
    def on_stage_started(self, stage: Stage): ...

    def on_stage_ended(self, stage: Stage, errors_count: int, was_interrupted: bool): ...

#
# Types
#

class NullProgressListener(ProgressListener):
    def on_stage_started(self, stage: Stage):
        pass

    def on_stage_ended(self, stage: Stage, errors_count: int, was_interrupted: bool):
        pass

class StageRuntimeReporter(ProgressListener):
    def __init__(self, is_verbose: bool, print_impl: Callable[[str], None] = print):
        self.__init_time_seconds         = default_timer()
        self.__stage_start_times_seconds = dict()
        self.__stage_end_times_seconds   = dict()
        self.__stage_errors_counts       = dict()
        self.__interrupted_stages        = set()
        self.__is_verbose                = is_verbose
        self.__print_impl                = print_impl

    @property
    def current_stage(self) -> Stage | None:
        unfinished_stages = self.__stage_start_times_seconds.keys() - self.__stage_end_times_seconds.keys()
        assert len(unfinished_stages) <= 1, "must not have multiple unfinished stages simultaneously"
        try:
            return next(iter(unfinished_stages))
        except StopIteration:
            return None

    def on_stage_started(self, stage: Stage):
        assert stage not in self.__stage_start_times_seconds
        self.__stage_start_times_seconds[stage] = default_timer()

    def on_stage_ended(self, stage: Stage, errors_count: int, was_interrupted: bool):
        assert stage in self.__stage_start_times_seconds and stage not in self.__stage_end_times_seconds
        assert stage not in self.__stage_errors_counts
        self.__stage_end_times_seconds[stage] = default_timer()
        self.__stage_errors_counts[stage]     = errors_count
        if was_interrupted:
            self.__interrupted_stages.add(stage)
        stage_verb_str = "interrupted after" if was_interrupted else "completed in"
        self.__print_impl(f"{stage.value} {stage_verb_str} {self.__stage_time_seconds(stage):.1f} seconds{self.__format_errors_count(' with {0}', errors_count)}")

    def log_summary(self):
        total_runtime_seconds = default_timer() - self.__init_time_seconds
        self.__print_impl(f"Total runtime: {total_runtime_seconds:.1f} seconds{', stages:' if self.__is_verbose else ''}")
        if not self.__is_verbose:
            return
        for stage in Stage:
            stage_time_seconds = self.__stage_time_seconds(stage)
            if stage_time_seconds is None:
                continue
            assert stage in self.__stage_errors_counts
            stage_stats_str = self.__format_stage_stats(stage_time_seconds, self.__stage_errors_counts[stage], stage in self.__interrupted_stages)
            self.__print_impl(f"- {stage.value:<27}{stage_stats_str}")

    def __stage_time_seconds(self, stage: Stage) -> float | None:
        if stage not in self.__stage_start_times_seconds or stage not in self.__stage_end_times_seconds:
            return None
        return self.__stage_end_times_seconds[stage] - self.__stage_start_times_seconds[stage]

    @staticmethod
    def __format_errors_count(format: str, errors_count: int) -> str:
        return format.format(f"{errors_count} error{'s' if errors_count != 1 else ''}") if errors_count > 0 else ""

    @staticmethod
    def __format_stage_stats(stage_time_seconds: float, errors_count: int, was_interrupted: bool) -> str:
        errors_count_str_template = ' ({0}, interrupted)' if was_interrupted else ' ({0})'
        errors_count_str          = StageRuntimeReporter.__format_errors_count(errors_count_str_template, errors_count)
        secondary_stats_str       = errors_count_str if len(errors_count_str) > 0 else (" (interrupted)" if was_interrupted else "")
        return f"{stage_time_seconds:.1f}s{secondary_stats_str}"



from enum import Enum
from sys import stderr

#
# Types
#

class _AnsiColor(str, Enum):
    DEFAULT = '\033[0m'
    RED     = '\033[31m'
    GREEN   = '\033[32m'
    YELLOW  = '\033[33m'

#
# Globals
#

_is_color_disabled = False

#
# Interface
#

def set_color_disabled(is_color_disabled: bool):
    global _is_color_disabled
    _is_color_disabled = is_color_disabled


def print_success(*args, **kwargs):
    _print_in_color(_AnsiColor.GREEN, *args, **kwargs)

def print_warning(*args, **kwargs):
    _print_in_color(_AnsiColor.YELLOW, *args, **dict({'file' : stderr}, **kwargs))

def print_error(*args, **kwargs):
    _print_in_color(_AnsiColor.RED, *args, **dict({'file' : stderr}, **kwargs))

#
# Service
#

def _print_in_color(color: _AnsiColor, *args, **kwargs):
    print(f"{'' if _is_color_disabled else color.value}{args[0]}{'' if _is_color_disabled else _AnsiColor.DEFAULT.value}", *(args[1:]), **kwargs)



#
# Types
#

class ModuleExecutionException(ImportError):
    def __init__(self, original_exception: BaseException):
        self.original_exception = original_exception

class InvalidVarsProcessorInterface(Exception):
    def __init__(self, vars_processor_module_locator: str, vars_processor_function_name: str):
        super().__init__(f"module '{vars_processor_module_locator}' does not expose the required {vars_processor_function_name}(vars: TemplateVariables) function")



import argparse
import traceback
import atexit
import signal
from pathlib import Path
from os import system
from enum import Enum
from sys import stdout, platform as sys_platform
from io import TextIOWrapper

from . import TemplateVariables, RenderConfig, RenderDirResult, load_vars_from_yaml_files, post_process_vars, render_dir
from .exceptions import ModuleExecutionException, InvalidVarsProcessorInterface
from .timing import StageRuntimeReporter
from ._logging import *

#
# Constants
#

_CLI_DESCRIPTION = """
Renders all Jinja templates in a directory into files in another directory, preserving the folder structure.
Templates must use UTF-8 (with or without BOM), rendered files will reflect their source templates' BOM or lack thereof.
"""

_VARS_PROCESSOR_MODULE_ARG   = '--vars-processor'
_VARS_PROCESSOR_FUNCTION_ARG = '--vars-processor-func'

_DEFAULT_VARS_PROCESSOR_FUNCTION_NAME = 'process_vars'

_SIGNAL_EXIT_CODE_OFFSET = 128

#
# Command line arguments
#

def _make_cli_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(prog='kannushi', description=_CLI_DESCRIPTION)

    parser.add_argument('source_path', metavar='SOURCE_PATH', type=Path, help='root directory containing Jinja templates')
    parser.add_argument('target_path', metavar='TARGET_PATH', type=Path, help='target root directory for rendered files')

    parser.add_argument('--skip', dest='skip_glob', metavar='SKIP_GLOB', type=str, help='glob for template files to skip when rendering (relative to SOURCE_PATH)')
    parser.add_argument('--vars', dest='vars_glob', metavar='VARS_YAML_GLOB', type=str, help='YAML file(s) containing template variable definitions')

    parser.add_argument(
        _VARS_PROCESSOR_MODULE_ARG, dest='vars_processor_module_locator', metavar='VARS_PROCESSOR_MODULE', type=str, help='Python file/module to use for variables dictionary post-processing')
    parser.add_argument(
        _VARS_PROCESSOR_FUNCTION_ARG, dest='vars_processor_function_name', metavar='VARS_PROCESSOR_FUNCTION', type=str, default=_DEFAULT_VARS_PROCESSOR_FUNCTION_NAME,
        help=f'single-parameter function in VARS_PROCESSOR_MODULE which vars dictionary will be passed to (defaults to {_DEFAULT_VARS_PROCESSOR_FUNCTION_NAME})'
    )

    parser.add_argument('--seed', dest='random_seed', metavar='RANDOM_SEED', type=int, help='RNG seed to use for any randomization')

    parser.add_argument('-j', '--jobs', dest='jobs_count', metavar='JOBS_COUNT', type=int, help='number of parallel jobs (defaults to the number of logical CPU cores)')

    parser.add_argument('-v', '--verbose', action='store_true', dest='is_verbose', help='output processed file paths to stdout')
    parser.add_argument('--no-color', action='store_true', dest='is_color_disabled', help='disable output coloring')

    return parser

#
# Types
#

class _MainExitCode(int, Enum):
    SUCCESS                = 0
    UNKNOWN_ERROR          = 1
    VARS_LOADING_FAILED    = 2
    VARS_PROCESSING_FAILED = 3
    JINJA_RENDER_ERRORS    = 4
    INTERRUPTED            = _SIGNAL_EXIT_CODE_OFFSET + signal.SIGINT

    @staticmethod
    def from_render_dir_result(render_dir_result: RenderDirResult) -> "_MainExitCode":
        if render_dir_result.was_interrupted:
            return _MainExitCode.INTERRUPTED
        if render_dir_result.errors_count > 0:
            return _MainExitCode.JINJA_RENDER_ERRORS
        return _MainExitCode.SUCCESS

#
# Service
#

def _make_render_config_from_args(args: argparse.Namespace) -> RenderConfig:
    return RenderConfig(
        source_path=args.source_path,
        target_path=args.target_path,
        skip_glob=args.skip_glob,
        random_seed=args.random_seed,
        requested_jobs_count=args.jobs_count,
        is_verbose=args.is_verbose,
        is_color_disabled=args.is_color_disabled,
    )

#
# Main
#

def main():
    parser = _make_cli_parser()
    args   = parser.parse_args()

    set_color_disabled(args.is_color_disabled)
    if not args.is_color_disabled and sys_platform == 'win32':
        system('color')

    config = _make_render_config_from_args(args)

    if isinstance(stdout, TextIOWrapper):
        stdout.reconfigure(line_buffering=True)

    stage_time_reporter = StageRuntimeReporter(config.is_verbose)
    atexit.register(lambda: stage_time_reporter.log_summary())

    try:
        vars = load_vars_from_yaml_files(args.vars_glob, config.effective_jobs_count, stage_time_reporter) if not args.vars_glob is None else TemplateVariables()
    except KeyboardInterrupt:
        print_warning('warning: Interrupted by the user')
        exit(_MainExitCode.INTERRUPTED)
    except BaseException as e:
        print_error('\n'.join(traceback.format_exception_only(e)))
        exit(_MainExitCode.VARS_LOADING_FAILED)

    if args.vars_processor_module_locator is not None:
        try:
            post_process_vars(vars, args.vars_processor_module_locator, args.vars_processor_function_name, stage_time_reporter)
        except KeyboardInterrupt:
            print_warning('warning: Interrupted by the user')
            exit(_MainExitCode.INTERRUPTED)
        except ModuleExecutionException as e:
            print_error('\n'.join(traceback.format_exception(e.original_exception)))
            print_error(f"error: Failed to load module {args.vars_processor_module_locator} due to the exception above")
            exit(_MainExitCode.VARS_PROCESSING_FAILED)
        except ImportError as e:
            print_error(f"error: {e}")
            print(f'hint: make sure a valid Python module name or .py file path is given via {_VARS_PROCESSOR_MODULE_ARG}')
            exit(_MainExitCode.VARS_PROCESSING_FAILED)
        except InvalidVarsProcessorInterface as e:
            print_error(f"error: {e}")
            exit(_MainExitCode.VARS_PROCESSING_FAILED)
        except BaseException as e:
            print_error('\n'.join(traceback.format_exception(e)))
            print_error(f"error: Failed to process variables using {args.vars_processor_module_locator} due to the exception above")
            exit(_MainExitCode.VARS_PROCESSING_FAILED)
    elif args.vars_processor_function_name != _DEFAULT_VARS_PROCESSOR_FUNCTION_NAME:
        print_warning(f"warning: Ignoring {_VARS_PROCESSOR_FUNCTION_ARG} in the absence of {_VARS_PROCESSOR_MODULE_ARG}")

    result = render_dir(config, vars, stage_time_reporter)
    if result.was_interrupted:
        print_warning(f"warning: Interrupted by the user ({result.skipped_count} template{'s' if result.skipped_count != 1 else ''} skipped)")
    if result.errors_count > 0:
        if config.is_verbose:
            failed_template_paths_str  = '\n'.join(map(str, result.failed_template_paths))
            print_error(f"error: The following {result.errors_count} template{'s' if result.errors_count != 1 else ''} failed to render (see individual errors above):")
            print_error(failed_template_paths_str)
        else:
            print_error(f"error: {result.errors_count} template{'s' if result.errors_count != 1 else ''} failed to render (see individual errors above)")
    elif not result.was_interrupted:
        assert result.is_successful
        print_success(f"All {result.rendered_templates_count} templates rendered without errors")
    exit(_MainExitCode.from_render_dir_result(result))



#
# Types
#

# A convenience wrapper around dict, allowing to acess values
# both by key via [] or as regular attributes via dot notation.
class TemplateVariables(dict):
     def __init__(self, vars: dict = {}):
         super().__init__(vars)
         self.__dict__ = self



import random
import signal
from pathlib import Path
from os import path, cpu_count
from dataclasses import dataclass, field
from typing import Callable, cast
from multiprocessing import Pool
from multiprocessing.pool import AsyncResult
from functools import cached_property, partial
from sys import stdout
from timeit import default_timer

from jinja2 import Environment, FileSystemLoader, StrictUndefined
from jinja2_error import ErrorExtension

from .timing import Stage, ProgressListener, NullProgressListener
from ._vars import TemplateVariables
from ._vars.loading import inject_service_var
from ._logging import set_color_disabled, print_warning, print_error

#
# Constants
#

_SOURCE_ENCODING = 'utf-8' # Treating both source and rendered content as regular UTF-8 handles BOM correctly.
_TARGET_ENCODING = 'utf-8' #

_TEMPLATE_EXTENSION = '.jinja'
_TEMPLATE_GLOB      = '**/*' + _TEMPLATE_EXTENSION

_TEMPLATE_PATH_VAR = '_template_path' # Name of the template variable to be set to the currently rendered template's path

_ASYNC_POLLING_INTERVAL_SECONDS = 0.5 # Primarily determines the reaction time to Ctrl-C (KeyboardInterrupt)

#
# Interface types
#

@dataclass
class RenderConfig:
    source_path:          Path
    target_path:          Path
    skip_glob:            str | None
    random_seed:          int | None
    requested_jobs_count: int | None
    is_verbose:           bool
    is_color_disabled:    bool

    @cached_property
    def effective_jobs_count(self) -> int:
        return self.__try_cap_jobs_count(self.requested_jobs_count) or cpu_count() or 1

    @staticmethod
    def __try_cap_jobs_count(jobs_count: int | None) -> int | None:
        if jobs_count is None:
            return None
        if (cpu_count_value := cpu_count()) is not None and jobs_count > cpu_count_value:
            print_warning(f"warning: Capping the number of jobs to {cpu_count_value} - the number of logical CPU cores", file=stdout)
            return cpu_count_value
        return jobs_count

@dataclass
class RenderDirResult:
    selected_templates_count: int        = 0
    rendered_templates_count: int        = 0
    failed_template_paths:    list[Path] = field(default_factory=list)
    was_interrupted:          bool       = False

    @property
    def errors_count(self) -> int:
        return len(self.failed_template_paths)

    @property
    def skipped_count(self) -> int:
        result = self.selected_templates_count - self.rendered_templates_count - self.errors_count
        assert result >= 0
        return result

    @property
    def is_successful(self) -> bool:
        return self.selected_templates_count == self.rendered_templates_count

#
# Service types
#

@dataclass
class _RenderTemplateResult:
    target_file_path:    Path
    render_time_seconds: float

#
# Globals
#

_jinja_env = None
_vars      = None

#
# Interface
#

def render_dir(config: RenderConfig, vars: TemplateVariables, progress_listener: ProgressListener = NullProgressListener()) -> RenderDirResult:
    templates_paths = config.source_path.glob(_TEMPLATE_GLOB)
    skipped_paths   = config.source_path.glob(config.skip_glob) if config.skip_glob is not None else []
    selected_paths  = [template_path for template_path in templates_paths if template_path not in skipped_paths]

    current_stage = None
    def change_stage(stage: Stage | None, current_stage_errors_count: int = 0, was_interrupted: bool = False):
        nonlocal current_stage
        if current_stage is not None:
            progress_listener.on_stage_ended(current_stage, current_stage_errors_count, was_interrupted)
        current_stage = stage
        if current_stage is not None:
            progress_listener.on_stage_started(current_stage)

    print(f'Initializing {config.effective_jobs_count} render processes...')
    change_stage(Stage.RENDER_POOL_INIT)

    def job_success_callback(template_result: _RenderTemplateResult):
        result.rendered_templates_count += 1
        if config.is_verbose:
            print(f'[{template_result.render_time_seconds:4.2f}s] {template_result.target_file_path}')

    def job_error_callback(template_path: Path, e: BaseException):
        result.failed_template_paths.append(template_path)
        (_, target_file_path) = _convert_template_path(config.source_path, config.target_path, template_path)
        print_error(f'[ERROR] {target_file_path}')
        print_error(f'\terror: {e}')

    result = RenderDirResult()
    result.selected_templates_count = len(selected_paths)
    with Pool(config.effective_jobs_count, _init_render_template_process, (config.source_path, vars, config.is_color_disabled)) as process_pool:
        def render_template_async(template_path) -> AsyncResult:
            return process_pool.apply_async(
                _render_template_job,
                (config, template_path),
                callback=job_success_callback,
                error_callback=cast(Callable[[BaseException], None], partial(job_error_callback, template_path))
            )
        try:
            async_results = [render_template_async(template_path) for template_path in selected_paths]
            process_pool.close()

            change_stage(Stage.JINJA_RENDER)
            print(f'Rendering {len(selected_paths)} templates in {config.effective_jobs_count} parallel jobs...')

            for async_result in async_results:
                while not async_result.ready():
                    async_result.wait(_ASYNC_POLLING_INTERVAL_SECONDS)

            process_pool.join()
        except KeyboardInterrupt:
            result.was_interrupted = True
            process_pool.terminate()

    change_stage(None, result.errors_count if current_stage == Stage.JINJA_RENDER else 0, result.was_interrupted)

    return result

#
# Service
#

def _init_render_template_process(source_path: Path, vars: TemplateVariables, is_color_disabled: bool):
    # Prevent Ctrl-C from raising KeyboardInterrupt in child processes
    signal.signal(signal.SIGINT, signal.SIG_IGN)

    global _jinja_env, _vars
    _jinja_env = Environment(
        loader=FileSystemLoader(source_path, encoding=_SOURCE_ENCODING),
        extensions=['jinja2.ext.do', ErrorExtension],
        autoescape=False,
        undefined=StrictUndefined
    )
    _vars = vars
    set_color_disabled(is_color_disabled)

def _render_template_job(config: RenderConfig, template_path: Path) -> _RenderTemplateResult:
    """This function is the entry point for individual template rendering jobs run in parallel"""

    assert isinstance(_jinja_env, Environment)
    assert isinstance(_vars, dict)

    render_start_time_seconds = default_timer()

    (template_name, target_file_path) = _convert_template_path(config.source_path, config.target_path, template_path)

    inject_service_var(_vars, _TEMPLATE_PATH_VAR, _replace_backslashes(template_path))
    try:
        random.seed(config.random_seed)
        _render_template(_jinja_env, template_name, target_file_path, _vars)
    finally:
        del _vars[_TEMPLATE_PATH_VAR]

    return _RenderTemplateResult(target_file_path, default_timer() - render_start_time_seconds)

def _render_template(jinja_env: Environment, template_name: str, target_file_path: Path, vars: dict):
    template = jinja_env.get_template(template_name)
    rendered_content = template.render(vars)
    target_file_path.parent.mkdir(exist_ok=True, parents=True)
    with open(target_file_path, 'w', encoding=_TARGET_ENCODING) as target_file:
        target_file.write(rendered_content)

def _convert_template_path(source_dir_path: Path, target_dir_path: Path, template_path: Path) -> tuple[str, Path]:
    template_name    = _source_template_path_to_name(source_dir_path, template_path)
    target_file_path = _template_name_to_target_file_path(target_dir_path, template_name)
    return (template_name, target_file_path)

def _source_template_path_to_name(source_dir_path: Path, template_path: Path) -> str:
    return _replace_backslashes(path.relpath(template_path, source_dir_path))

def _replace_backslashes(path_str: str | Path) -> str:
    return str(path_str).replace('\\', '/')

def _template_name_to_target_file_path(target_dir_path: Path, template_name: str) -> Path:
    assert template_name.endswith(_TEMPLATE_EXTENSION)
    return target_dir_path / template_name[:-len(_TEMPLATE_EXTENSION)]
